/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */
package org.greenstone.gatherer.gems;

import java.io.*;
import java.util.*;
import org.greenstone.gatherer.Configuration;
import org.greenstone.gatherer.cdm.CommandTokenizer;
import org.greenstone.gatherer.util.ArrayTools;
import org.greenstone.gatherer.util.StaticStrings;
import org.greenstone.gatherer.util.Utility;
import org.greenstone.gatherer.util.XMLTools;
import org.w3c.dom.*;
/** This class contains a plethora of methods associated with handling the content of <strong>MetadataSet</strong>s and the <strong>Element</strong>s within. For example this is where you will find methods for comparing various parts of two <strong>MetadataSet</strong>s for equality. It also has methods for extracting common and useful data from <strong>Element</strong>s such as the AssignedValue nodes.
 * @author John Thompson, Greenstone Digital Libraries
 * @version 2.3
 */
public class MSMUtils {
    static public int NONE = 0;     //An element of the enumeration of type filter.
    static public int VALUES = 1;   //An element of the enumeration of type filter.
    static public int ALIASES = 2;  //An element of the enumeration of type filter.
    static public int BOTH = 3;     //An element of the enumeration of type filter.
    static public char NS_SEP= '.'; //The character used to separate name space from metadata element.
    static public char SF_SEP= '^'; //The character used to separate subfields from metadata element.


    /** Method to add an attribute element to the given element. This method makes use of the language_dependent attribute of the document to not only determine if the attribute is language dependent, but also to see whether a Language element should be created if doesn't already exist.
     * @param element_element the Element to add the attribute element to
     * @param attribute_name_str the name of the new attribute to add as a String
     * @param language_code_str the two letter code String of the language this attribute is to be added as
     * @param value_str the String to be assigned as the attribute elements value
     * @see org.greenstone.gatherer.msm.MSMUtils#isAttributeLanguageDependent
     * @see org.greenstone.gatherer.msm.MSMUtils#setValue(Element, String)
     * @see org.greenstone.gatherer.util.StaticStrings#ATTRIBUTE_ELEMENT
     * @see org.greenstone.gatherer.util.StaticStrings#CODE_ATTRIBUTE
     * @see org.greenstone.gatherer.util.StaticStrings#LANGUAGE_ATTRIBUTE
     * @see org.greenstone.gatherer.util.StaticStrings#LANGUAGE_ELEMENT
     * @see org.greenstone.gatherer.util.StaticStrings#NAME_ATTRIBUTE
     */ 
    static public void addElementAttribute(Element element_element, String attribute_name_str, String language_code_str, String value_str) {
	Document document = element_element.getOwnerDocument();
	// Create the basic new attribute (everything except language attribute)
	Element attribute_element = document.createElement(StaticStrings.ATTRIBUTE_ELEMENT);
	attribute_element.setAttribute(StaticStrings.NAME_ATTRIBUTE, attribute_name_str);
	MSMUtils.setValue(attribute_element, value_str);

	// Start off by determining if we have to add this node in the new multilingual optimized way
	if(isAttributeLanguageDependent(document, attribute_name_str)) {
	    boolean found = false;
	    // Try to retrieve a language element for the given language code
	    //NodeList language_elements = element_element.getElementsByTagName(StaticStrings.LANGUAGE_ELEMENT);
	    ArrayList language_elements = XMLTools.getChildElementsByTagName(element_element, StaticStrings.LANGUAGE_ELEMENT);
	    for(int i = 0; i < language_elements.size(); i++) {
		Element language_element = (Element) language_elements.get(i);
		if(language_element.getAttribute(StaticStrings.CODE_ATTRIBUTE).equals(language_code_str)) {
		    found = true;
		    // Add attribute
		    language_element.appendChild(attribute_element);
		}
		language_element = null;
	    }
	    language_elements = null;
	    // If it still hasn't been found, then add it
	    if(!found) {
		Element language_element = document.createElement(StaticStrings.LANGUAGE_ELEMENT);
		language_element.setAttribute(StaticStrings.CODE_ATTRIBUTE, language_code_str);
		element_element.appendChild(language_element);
		// Add attribute
		language_element.appendChild(attribute_element);
		language_element = null;
	    }
	}
	// Just add the attribute the old fashioned way
	else {
	    attribute_element.setAttribute(StaticStrings.LANGUAGE_ATTRIBUTE, language_code_str);
	    element_element.appendChild(attribute_element);
	}
	// Clean up
	attribute_element = null;
	document = null;
    }


    /** Retrieve all of the attributes for the given element as a tree set. Note that this requires significant manipulation if the source is a multilingual optimized metadata set.
     * @param element the Element whose attributes we wish to catalog
     * @return a TreeSet of the attributes sorted by their natural ordering
     * @see org.greenstone.gatherer.msm.MSMUtils#getValue(Node)
     * @see org.greenstone.gatherer.util.StaticStrings#ATTRIBUTE_ELEMENT
     * @see org.greenstone.gatherer.util.StaticStrings#CODE_ATTRIBUTE
     * @see org.greenstone.gatherer.util.StaticStrings#LANGUAGE_ATTRIBUTE
     * @see org.greenstone.gatherer.util.StaticStrings#LANGUAGE_ELEMENT
     * @see org.greenstone.gatherer.util.StaticStrings#NAME_ATTRIBUTE
     */
    static public TreeSet getAttributes(Element element) {
	TreeSet attribute_tree = new TreeSet();

        GEMSLanguageManager gemsLangManager = new GEMSLanguageManager(Configuration.gsdl_path + File.separator + "gli" + File.separator + "classes" + File.separator + "xml" + File.separator + "languages.xml");
        
        String[] enabled_langs = gemsLangManager.getEnabledLanguageCodesToArray();
                       
	for(Node node = element.getFirstChild(); node != null; node = node.getNextSibling()) {
	    if(node instanceof Element) {
		Element some_element = (Element) node;
		String some_element_name = some_element.getNodeName();
		if(some_element_name.equals(StaticStrings.ATTRIBUTE_ELEMENT)) {
		    attribute_tree.add(new Attribute(some_element.getAttribute(StaticStrings.NAME_ATTRIBUTE), some_element.getAttribute(StaticStrings.LANGUAGE_ATTRIBUTE), MSMUtils.getValue(some_element)));
                    // System.out.println(some_element_name +  "<- some_element nodes' value");
                }
		else if(some_element_name.equals(StaticStrings.LANGUAGE_ELEMENT)) {
		    String language_code = some_element.getAttribute(StaticStrings.CODE_ATTRIBUTE);
		    NodeList attribute_elements = some_element.getElementsByTagName(StaticStrings.ATTRIBUTE_ELEMENT);
                    
		    for(int i = 0; i < attribute_elements.getLength(); i++) {
			Element attribute_element = (Element) attribute_elements.item(i);
                     
                            for(int k = 0; k < enabled_langs.length; k++){
                                
                               if(enabled_langs[k].trim().compareTo(some_element.getAttributes().item(0).getNodeValue().toString()) == 0){
				   attribute_tree.add(new Attribute(attribute_element.getAttribute(StaticStrings.NAME_ATTRIBUTE), language_code, MSMUtils.getValue(attribute_element)));
				   attribute_element = null;
				   
				   break;
                               }
                            }
			    attribute_element = null;
		   } 
		    attribute_elements = null;
		    language_code = null;
		}
		some_element_name = null;
		some_element = null;
	    }
	}
	return attribute_tree;
    }


    /** Method to construct an elements description by retrieving the correct attribute.
     * @param element the Element whose name we wish to retrieve
     * @return a String which is the elements description, or an empty string if no description exists
     * @see org.greenstone.gatherer.msm.MSMUtils#getElementAttribute
     * @see org.greenstone.gatherer.util.StaticStrings#COMMENT_VALUE
     * @see org.greenstone.gatherer.util.StaticStrings#DEFINITION_VALUE
     * @see org.greenstone.gatherer.util.StaticStrings#EMPTY_STR
     * @see org.greenstone.gatherer.util.StaticStrings#SPACE_CHARACTER
     */
    static public String getDescription(Element element) {
	String language_code_str = Configuration.getLanguage();
	StringBuffer description = new StringBuffer(StaticStrings.EMPTY_STR);
	description.append(getElementAttribute(element, StaticStrings.DEFINITION_VALUE, language_code_str));
	if(description.length() > 0) {
	    description.append(StaticStrings.SPACE_CHARACTER);
	}
	description.append(getElementAttribute(element, StaticStrings.COMMENT_VALUE, language_code_str));
	language_code_str = null;
	return description.toString();
    }


    /** Retrieve the value for the requested attribute in the required language. Once again this method must be aware of the differences between the old metadata sets and the new multilingual optimized ones.
     * @param element_element the Element whose attributes we are searching through
     * @param attribute_name_str the name of the desired attribute as a String
     * @param language_code_str the two letter code String indicating the desired language
     * @see org.greenstone.gatherer.msm.MSMUtils#getValue
     * @see org.greenstone.gatherer.msm.MSMUtils#isAttributeLanguageDependent
     * @see org.greenstone.gatherer.util.StaticStrings#ATTRIBUTE_ELEMENT
     * @see org.greenstone.gatherer.util.StaticStrings#CODE_ATTRIBUTE
     * @see org.greenstone.gatherer.util.StaticStrings#EMPTY_STR
     * @see org.greenstone.gatherer.util.StaticStrings#LANGUAGE_ATTRIBUTE
     * @see org.greenstone.gatherer.util.StaticStrings#LANGUAGE_ELEMENT
     * @see org.greenstone.gatherer.util.StaticStrings#NAME_ATTRIBUTE
     */
    static private String getElementAttribute(Element element_element, String attribute_name_str, String language_code_str) {
	boolean found = false;
	String result = StaticStrings.EMPTY_STR;
	Document document = element_element.getOwnerDocument();
	// Determine if the attribute is language specific
	if(isAttributeLanguageDependent(document, attribute_name_str)) {
	    NodeList language_elements = element_element.getElementsByTagName(StaticStrings.LANGUAGE_ELEMENT);
	    for(int i = 0; !found && i < language_elements.getLength(); i++) {
		Element language_element = (Element) language_elements.item(i);
		if(language_element.getAttribute(StaticStrings.CODE_ATTRIBUTE).equals(language_code_str)) {
		    NodeList attribute_elements = language_element.getElementsByTagName(StaticStrings.ATTRIBUTE_ELEMENT);
		    for(int j = 0; !found && j < attribute_elements.getLength(); j++) {
			Element attribute_element = (Element) attribute_elements.item(j);
			if(attribute_element.getAttribute(StaticStrings.NAME_ATTRIBUTE).equals(attribute_name_str)) {
			    found = true;
			    result = MSMUtils.getValue(attribute_element);
			}
			attribute_element = null;
		    }
		    attribute_elements = null;
		}
		language_element = null;
	    }
	    language_elements = null;
	}
	else {
	    boolean first_match = false;
	    NodeList attribute_elements = element_element.getElementsByTagName(StaticStrings.ATTRIBUTE_ELEMENT);
	    for(int k = 0; !found && k < attribute_elements.getLength(); k++) {
		Element attribute_element = (Element) attribute_elements.item(k);
		// We don't want to consider those attributes found inside language elements
		if(attribute_element.getParentNode() == element_element) {
		    ///ystem.err.println("First level");
		    String target_name_str = attribute_element.getAttribute(StaticStrings.NAME_ATTRIBUTE);
		    String target_language_str = attribute_element.getAttribute(StaticStrings.LANGUAGE_ATTRIBUTE);
		    ///ystem.err.println("Does " + target_name_str + " equal " + attribute_name_str + "?");
		    if(attribute_name_str.equals(target_name_str)) {
			///ystem.err.println("Does " + target_language_str + " equal " + language_code_str + "?");
			if(language_code_str.equals(target_language_str)) {
			    ///ystem.err.println("Perfect match!");
			    found = true;
			    result = MSMUtils.getValue(attribute_element);
			}
			else if((result == StaticStrings.EMPTY_STR || first_match) && isLegacyMDS(document)) {
			    ///ystem.err.println("Legacy MDS");
			    // Special case for old style documents, where the english match is good enough
			    if(target_language_str.equals(StaticStrings.ENGLISH_LANGUAGE_STR)) {
				///ystem.err.println("English plate.");
				result = MSMUtils.getValue(attribute_element);
			    }
			    // Super special case where the first match is better than nothing
			    else if(result == StaticStrings.EMPTY_STR && !first_match) {
				///ystem.err.println("First match.");
				first_match = true;
				result = MSMUtils.getValue(attribute_element);
			    }
			}
		    }
		    target_language_str = null;
		    target_name_str = null;
		}
		attribute_element = null;
	    }
	    attribute_elements = null;
	}
	document = null;
	return result;
    }



    /** Method to construct an elements fully qualified name. Note that this is different from a nodes identifier. Think of name as a short, unique reference to a metadata element, whereas identifier can be much longer, language specific and non-unique.
     * @param element An <strong>Element</strong> whose name we are interested in.
     * @return A <strong>String</strong> representing this given elements fully namespace qualified name.
     */
    static final public String getFullName(Element element) {
	return getFullName(element, "");

    }


    /** Method to construct an elements fully qualified name. Note that this is different from a nodes identifier. Think of name as a short, unique reference to a metadata element, whereas identifier can be much longer, language specific and non-unique.
     * @param element An <strong>Element</strong> whose name we are interested in.
     * @return A <strong>String</strong> representing this given elements fully namespace qualified name.
     */
    static final public String getFullName(Element element, String namespace) {
	StringBuffer name_buffer = new StringBuffer();
	if(element == null) {
	    return "Error";
	}
	// First get the root node.
	Document document = element.getOwnerDocument();
	Element root = document.getDocumentElement();
	document = null;
	// Retrieve this elements name
	name_buffer.append(element.getAttribute("name"));

	// Now we check if element has a parent node, other than root. If so we begin building up the full name
	//For some reason this doesn't work in specific cases of subelements.
	Element parent_element = (Element) element.getParentNode();
	while(parent_element != null && parent_element != root) { 
	    name_buffer.insert(0, SF_SEP);
	    name_buffer.insert(0, parent_element.getAttribute("name"));
	    parent_element = (Element)parent_element.getParentNode();
	}
	parent_element = null;
	// Finally insert the namespace and we are all done.
	if(root != null) {
	    namespace = root.getAttribute("namespace");
	} 
	root = null;
	// If no root, or no namespace found, assume its extracted (at least then they can't edit it)
	if(namespace == null || namespace.equals("")) {
	    namespace = Utility.EXTRACTED_METADATA_NAMESPACE;
	}
	name_buffer.insert(0, NS_SEP);
	name_buffer.insert(0, namespace);
	namespace = null;
	return name_buffer.toString();
    } // static public String getFullName(Element element)


    /**
       Method to get an element's name, not including its namespace.
       @param element an <strong>Element</strong> whose name we are interested in.
       @return A <strong>Strong</strong> representing the given elements name
       @author Matthew Whyte
       @date last modified: 21/01/04
     */
    static final public String getNameOnly(Element element)
    {
	StringBuffer name_buffer = new StringBuffer();
	if(element == null) {
	    return "Error";
	}
	// Retrieve this elements name
	name_buffer.append(element.getAttribute("name"));

	return name_buffer.toString();
    }


    /** Method to construct an elements name (sic identifier) by retrieving the correct attribute, language specific.
     * @param element the Element whose name we wish to retrieve
     * @return a String which is the elements identifier, or an empty string if no identifier exists
     * @see org.greenstone.gatherer.msm.MSMUtils#getElementAttribute
     * @see org.greenstone.gatherer.util.StaticStrings#IDENTIFIER_VALUE
     * @see org.greenstone.gatherer.util.StaticStrings#NAME_ATTRIBUTE
     */
    static final public String getIdentifier(Element element) {
	String identifier = getElementAttribute(element, StaticStrings.IDENTIFIER_VALUE, Configuration.getLanguage());
	// Failing the above we return the nodes name instead.
	if(identifier == null || identifier.length() == 0) {
	    identifier = element.getAttribute(StaticStrings.NAME_ATTRIBUTE);
	}

	return identifier;
    }


    /** Method to retrieve from the node given, a certain child node with the specified name.
     * @param parent The <strong>Node</strong> whose children should be searched.
     * @param name The required nodes name as a <strong>String</strong>.
     * @return The requested <strong>Node</strong> if it is found, <i>null</i> otherwise.
     */ 
    static final public Node getNodeFromNamed(Node parent, String name) {
	Node child = null;
	for(Node i = parent.getFirstChild(); i != null && child == null; 
	    i = i.getNextSibling()) {
	    if(i.getNodeName().equals(name)) {
		child = i;
	    }
	}
	return child;
    }


    /** Look for the occurances 'field' of the element and return it if found.
     * @return An <i>int</i> which matches the number in the occurances attribute of the element, or 0 if no such attribute.
     */
    static final public int getOccurances(Element element) {
	int count = 0;
	String number = null;
	if((number = element.getAttribute("occurances")) != null) {
	    try {
		count = Integer.parseInt(number);
	    }
	    catch(Exception error) {
		count = 0;
	    }
	}
	return count;
    }	 


    /** Method to retrieve the value of a given node (not the assigned values tree!).
     * @param element The <strong>Element</strong> whose value we wish to find.
     * @return The value found as a <strong>String</strong>, or <i>null</i> if this element has no value.
     */
    static final public String getValue(Node element) {
	// If we've been given a subject node first retrieve its value node.
	if(element.getNodeName().equals("Subject")) {
	    element = getNodeFromNamed(element, "Value");
	}
	// If we've got a value node, then reconstruct the text. Remember that DOM will split text over 256 characters into several text nodes
	if(element != null && element.hasChildNodes()) {
	   StringBuffer text_buffer = new StringBuffer();
	   NodeList text_nodes = element.getChildNodes();
	   for(int i = 0; i < text_nodes.getLength(); i++) {
	      Node possible_text = text_nodes.item(i);
	      if(possible_text.getNodeName().equals(StaticStrings.TEXT_NODE)) {
		 text_buffer.append(possible_text.getNodeValue());
	      }
	   }
	   return text_buffer.toString();
	}
	return "";
    }


    /** Determine if the named attribute is language specific for this collection. This information is found in a DOM attribute of the document element, as a comma separated list of attribute names.
     * @param document the Document for which we wish to check the language requirements
     * @param attribute_name_str the name of the attribute we a testing as a String
     * @see org.greenstone.gatherer.util.StaticStrings#LANGUAGEDEPENDENT_ATTRIBUTE
     * Modified by Matthew Whyte 4/02/05
     */
    static private boolean isAttributeLanguageDependent(Document document, String attribute_name_str)
    {
	//gives a list of comma seperated language specific attributes.
	String language_specific_attributes = document.getDocumentElement().getAttribute(StaticStrings.LANGUAGEDEPENDENT_ATTRIBUTE).toLowerCase();
	//System.err.println("lang_specific: " + language_specific_attributes); //debug

	//Seperate out language_specific_attributes
	String[] attributes_split = language_specific_attributes.split(",");

	//Compare each to attribute_name_str
	for(int i = 0; i < attributes_split.length; i++)
	    {
		if(attributes_split[i].trim().compareTo(attribute_name_str.toLowerCase()) == 0)
		    return true;
	    }
	return false;
    }


    /** A method to remove a specific attribute element from an element. This attribute must match in name, language and in value before being removed. Note that this method supports both legacy and multilingual optimized versions of the mds.
     * @param element_element the Element which represent the metadata element we are altering
     * @param attribute_name_str the name of the attribute to remove as a String
     * @param language_code_str the language code we must match as a String
     * @param value_str the value String which also must match before we remove anything
     * @return true if the desired attribute was successfully found and removed, false otherwise
     * @see org.greenstone.gatherer.msm.MSMUtils#isAttributeLanguageDependent
     * @see org.greenstone.gatherer.msm.MSMUtils#getValue(Node)
     * @see org.greenstone.gatherer.util.StaticStrings#ATTRIBUTE_ELEMENT
     * @see org.greenstone.gatherer.util.StaticStrings#CODE_ATTRIBUTE
     * @see org.greenstone.gatherer.util.StaticStrings#LANGUAGE_ATTRIBUTE
     * @see org.greenstone.gatherer.util.StaticStrings#LANGUAGE_ELEMENT
     * @see org.greenstone.gatherer.util.StaticStrings#NAME_ATTRIBUTE
     */
    static public boolean removeElementAttribute(Element element_element, String attribute_name_str, String language_code_str, String value_str) {
	//Determine the if this is one of the language specific attributes.
	//This can sometimes give a false result, if an attribute is made language specific half way through creation of a metadata set (or vice versa).
	if(isAttributeLanguageDependent(element_element.getOwnerDocument(), attribute_name_str))
	    {
		/*
		Multilingual Optimized version, eg:
		<Language code="en">
		    <Attribute name="identifier">second</Attribute>
		</Language>
		*/
		return removeElementAttributeNew(element_element, attribute_name_str, language_code_str, value_str, false);
	    }
	else
	    {
		/*
		The old way, eg:
		<Attribute language="en" name="identifier">second</Attribute>
		*/
		return removeElementAttributeOld(element_element, attribute_name_str, language_code_str, value_str, false);
	    }
	
    }
    static final private boolean removeElementAttributeNew(Element element_element, String attribute_name_str, String language_code_str, String value_str, boolean run_once)
    {
	// Retrieve the language elements, and determine the correct one
	NodeList language_elements = element_element.getElementsByTagName(StaticStrings.LANGUAGE_ELEMENT);
	for(int i = 0; i < language_elements.getLength(); i++) {
	    Element language_element = (Element) language_elements.item(i);
	    if(language_element.getAttribute(StaticStrings.CODE_ATTRIBUTE).equalsIgnoreCase(language_code_str)) {
		NodeList attribute_elements = language_element.getElementsByTagName(StaticStrings.ATTRIBUTE_ELEMENT);
		for(int j = 0; j < attribute_elements.getLength(); j++) {
		    Element attribute_element = (Element) attribute_elements.item(j);
		    String target_name_str = attribute_element.getAttribute(StaticStrings.NAME_ATTRIBUTE);
		    String target_value_str = MSMUtils.getValue(attribute_element);
		    if(attribute_name_str.equals(target_name_str) && value_str.equals(target_value_str)) {
			language_element.removeChild(attribute_element);
			if(attribute_elements.getLength() == 0) {
			    element_element.removeChild(language_element);
			}
			target_value_str = null;
			target_name_str = null;
			attribute_element = null;
			attribute_elements = null;
			language_element = null;
			language_elements = null;
			return true;
		    }
		    target_value_str = null;
		    target_name_str = null;
		    attribute_element = null;			
		}
		attribute_elements = null;
	    }
	    language_element = null;
	}
	language_elements = null;
	// Not found. Try the old way before failing.
	if(!run_once)
	    return removeElementAttributeOld(element_element, attribute_name_str, language_code_str, value_str, true);
	return false;
    }
    static final private boolean removeElementAttributeOld(Element element_element, String attribute_name_str, String language_code_str, String value_str, boolean run_once)
    {
	// Find the attribute to remove
	NodeList attribute_elements = element_element.getElementsByTagName(StaticStrings.ATTRIBUTE_ELEMENT);
	for (int k = 0; k < attribute_elements.getLength(); k++) {
	    Element attribute_element = (Element) attribute_elements.item(k);
	    // Remember to ignore any attributes that live within nested language elements
	    if (attribute_element.getParentNode() == element_element && attribute_element.getAttribute(StaticStrings.NAME_ATTRIBUTE).equals(attribute_name_str) && attribute_element.getAttribute(StaticStrings.LANGUAGE_ATTRIBUTE).equalsIgnoreCase(language_code_str) && MSMUtils.getValue(attribute_element).equals(value_str)) {
		// Match found, so remove the attribute node and return
		element_element.removeChild(attribute_element);
		attribute_element = null;
		attribute_elements = null;
		return true;
	    }
	    attribute_element = null;
	}
	attribute_elements = null;
	// No match found. Try the new way before failing.
	if(!run_once)
	    return removeElementAttributeNew(element_element, attribute_name_str, language_code_str, value_str, true);
	return false;
    }


    /** can only be used to set the english value */
    static final public void setIdentifier(Node element, String value) {
	// Get the 'identifier' Element
	for(Node node = element.getFirstChild(); node != null;
	    node = node.getNextSibling()) {
	    if(node.getNodeName().equals("Attribute")) {
		Element target = (Element)node;
		if(target.getAttribute("name").equals("identifier") && target.getAttribute("language").equals("en")) {
		    Node text = target.getFirstChild();
		    text.setNodeValue(value);
		    break;
		}
	    }
	}
    }


    /** Set the value of the element attribute occurances.
     * @param element The <strong>Element</strong> to change.
     * @param value The value to change by as an <i>int</i>.
     */
    static final public void setOccurance(Element element, int value) {
	Integer new_value = new Integer(getOccurances(element) + value);
	element.setAttribute("occurances", new_value.toString());
    }


    /** Set the #text node value of some element.
     * @param element the Element whose value we wish to set
     * @param value the new value for the element as a String
     */
    static final public void setValue(Element element, String value) {
	// Remove any existing child node(s)
	clear(element);
	// Add new text node.
	if (value != null) {
	    element.appendChild(element.getOwnerDocument().createTextNode(value));
	}
    }


    /** This method also traverses the tree, but this one is used to gather all the values and aliases at once, and to 'prune' the tree if necessary.
     * @param current The current root <strong>Node</strong> of this AssignedValues tree subtree.
     * @param return_filter This <i>int</i> specifies what nodes from the tree should be returned, where;<br>VALUES  = return values only<br>ALIASES = return aliases only<br>BOTH    = return both values and aliases<br>NONE    = return nothing.
     * @param remove_leaves A leaf node is a subject that contains no child subjects. If this <i>boolean</i> is set to <i>true</i> then the leaf nodes will be removed from the tree. 
     * @return A <strong>Node[]</strong> containing whatever values or aliases have been found during the trees traversal.
     */
    static final public Node[] traverseTree(Node current, int return_filter, boolean remove_leaves) {
	Node leaves[] = null;
	String name = current.getNodeName();
	if(name.equals("Value") && (return_filter == VALUES || return_filter == BOTH)) {
	    leaves = ArrayTools.add(leaves, current);
	}
	else if(name.equals("Alias") && (return_filter == ALIASES || return_filter == BOTH)) {
	    leaves = ArrayTools.add(leaves, current);
	}
	else if(name.equals("Subject")) {
	    boolean has_subject_child = false;
	    Node children[] = ArrayTools.nodeListToNodeArray(current.getChildNodes());
	    for(int i = 0; i < children.length; i++) {
		if(children[i].getNodeName().equals("Subject")) {
		    has_subject_child = true;
		}
		leaves = ArrayTools.add(leaves, traverseTree(children[i], return_filter, remove_leaves));
	    }
	    if(!has_subject_child && remove_leaves) {
		Node parent = current.getParentNode();
		parent.removeChild(current);
	    }
	}
	else if(name.equals("AssignedValues")) {
	    Node children[] = ArrayTools.nodeListToNodeArray(current.getChildNodes());
	    for(int i = 0; i < children.length; i++) {
		leaves = ArrayTools.add(leaves, traverseTree(children[i], return_filter, remove_leaves));
	    }
	}															  
	return leaves;
    }


    /** Determine if the given document is a legacy MDS or a new multilingual one. The easiest way to tell is whether there is a language_dependant attribute in the document element.
     * @param document the Document to test
     * @return true if this is an old mds, false otherwise
     * @see org.greenstone.gatherer.util.StaticStrings#LANGUAGEDEPENDANT_ATTRIBUTE
     * @see org.greenstone.gatherer.util.StaticStrings#EMPTY_STR
     */
    static private boolean isLegacyMDS(Document document) {
	///ystem.err.println("isLegacyMDS(): l_d = " + document.getDocumentElement().getAttribute(StaticStrings.LANGUAGEDEPENDANT_ATTRIBUTE));
	return (document.getDocumentElement().getAttribute(StaticStrings.LANGUAGEDEPENDENT_ATTRIBUTE)).equals(StaticStrings.EMPTY_STR);
    }


    /** Remove all of the child nodes from a certain node. */
    static final public void clear(Node parent) {
	while(parent.hasChildNodes()) {
	    parent.removeChild(parent.getFirstChild());
	}
    }
}
